#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Reusable Caching Primitives (:mod:`tendril.utils.www.caching`)
==============================================================

TODO Some introduction

"""


import os
import time
import codecs
import tempfile

from fs import open_fs
from fs.copy import copy_file
from fs.osfs import OSFS
from tendril.utils.fsutils import temp_fs
from tendril.config import INSTANCE_CACHE

from .status import is_connected

from tendril.utils import log
logger = log.get_logger(__name__, log.WARNING)

WWW_CACHE = os.path.join(INSTANCE_CACHE, 'soupcache')


class CacheBase(object):
    def __init__(self, cache_dir=WWW_CACHE):
        """
        This class implements a simple filesystem cache which can be used
        to create and obtain from various cached requests from internet
        resources.

        The cache is stored in the folder defined by ``cache_dir``, with a
        filename constructed by the :func:`_get_filepath` function.

        If the cache's :func:`_accessor` function is called with the
        ``getcpath`` attribute set to True, only the path to a (valid) file
        in the cache filesystem is returned, and opening and reading
        the file is left to the caller. This hook is provided to help deal
        with file encoding on a somewhat case-by-case basis, until the
        overall encoding problems can be ironed out.
        """
        self.cache_fs = open_fs(cache_dir, create=True)

    def _get_filepath(self, *args, **kwargs):
        """
        Given the parameters necessary to obtain the resource in normal
        circumstances, return a hash which is usable as the filename for
        the resource in the cache.

        The filename must be unique for every resource, and filename
        generation must be deterministic and repeatable.

        Must be implemented in every subclass.
        """
        raise NotImplementedError

    def _get_fresh_content(self, *args, **kwargs):
        """
        Given the parameters necessary to obtain the resource in normal
        circumstances, obtain the content of the resource from the source.

        Must be implemented in every subclass.
        """
        raise NotImplementedError

    @staticmethod
    def _serialize(response):
        """
        Given a response (as returned by :func:`_get_fresh_content`), convert
        it into a string which can be stored in a file. Use this function to
        serialize structured responses when needed.

        Unless overridden by the subclass, this function simply returns the
        response unaltered.

        The actions of this function should be reversed by
        :func:`_deserialize`.
        """
        return response

    @staticmethod
    def _deserialize(filecontent):
        """
        Given the contents of a cache file, reconstruct the original response
        in the original format (as returned by :func:`_get_fresh_content`).
        Use this function to deserialize cache files for structured responses
        when needed.

        Unless overridden by the subclass, this function simply returns the
        file content unaltered.

        The actions of this function should be reversed by
        :func:`_serialize`.
        """
        return filecontent

    def _cached_exists(self, filepath):
        return self.cache_fs.exists(filepath)

    def _is_cache_fresh(self, filepath, max_age):
        """
        Given the path to a file in the cache and the maximum age for the
        cache content to be considered fresh, returns (boolean) whether or
        not the cache contains a fresh copy of the response.

        :param filepath: Path to the filename in the cache corresponding to
                         the request, as returned by :func:`_get_filepath`.
        :param max_age: Maximum age of fresh content, in seconds.

        """
        if self._cached_exists(filepath):
            tn = int(time.time())
            tc = int(time.mktime(
                self.cache_fs.getinfo(filepath)['modified_time'].timetuple())
            )
            if tn - tc < max_age:
                return True
        return False

    def _accessor(self, max_age, getcpath=False, *args, **kwargs):
        """
        The primary accessor for the cache instance. Each subclass should
        provide a function which behaves similarly to that of the original,
        un-cached version of the resource getter. That function should adapt
        the parameters provided to it into the form needed for this one, and
        let this function maintain the cached responses and handle retrieval
        of the response.

        If the module's :data:`_internet_connected` is set to False, the
        cached value is returned regardless.

        """
        filepath = self._get_filepath(*args, **kwargs)
        send_cached = False
        if not is_connected() and self._cached_exists(filepath):
            send_cached = True
        if self._is_cache_fresh(filepath, max_age):
            logger.debug("Cache HIT")
            send_cached = True
        if send_cached is True:
            if getcpath is False:
                try:
                    filecontent = self.cache_fs.open(filepath, 'rb').read()
                    return self._deserialize(filecontent)
                except UnicodeDecodeError:
                    # TODO This requires the cache_fs to be a local
                    # filesystem. This may not be very nice. A way
                    # to hook codecs upto to pyfilesystems would be better
                    with codecs.open(
                            self.cache_fs.getsyspath(filepath),
                            encoding='utf-8') as f:
                        filecontent = f.read()
                        return self._deserialize(filecontent)
            else:
                return self.cache_fs.getsyspath(filepath)

        logger.debug("Cache MISS")
        data = self._get_fresh_content(*args, **kwargs)

        sdata = self._serialize(data)
        fd, temppath = tempfile.mkstemp()
        fp = os.fdopen(fd, 'wb')
        fp.write(sdata)
        fp.close()
        logger.debug("Creating new cache entry")
        # This can be pretty expensive if the move is across a real
        # filesystem boundary. We should instead use a temporary file
        # in the cache_fs itself
        try:
            copy_file(temp_fs, temp_fs.unsyspath(temppath),
                      self.cache_fs, filepath)
            if isinstance(self.cache_fs, OSFS):
                # TODO Refine permissions
                os.chmod(self.cache_fs.getsyspath(filepath), 0o666)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:  # noqa
            logger.warning("Unable to write cache file "
                           "{0}".format(filepath))

        if getcpath is False:
            return data
        else:
            return self.cache_fs.getsyspath(filepath)





import ssl
from tendril.config import CA_BUNDLE
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


ssl_context = ssl.create_default_context()

if CA_BUNDLE:
    logger.info(f"Using custom CA Bundle at '{CA_BUNDLE}'")
    ssl_context.load_verify_locations(cafile=CA_BUNDLE)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Docstring for test_utils_www
"""

import six
from tendril.utils.www import redirectcache
from tendril.utils.www import bare
from hashlib import md5
import pytest
from six.moves.urllib.error import HTTPError, URLError
redirectcache.DUMP_REDIR_CACHE_ON_EXIT = False


def test_cached_fetcher():
    test_url = 'http://www.google.com'
    if six.PY3:
        filepath = md5(test_url.encode('utf-8')).hexdigest()
    else:
        filepath = md5(test_url).hexdigest()
    fs = bare.cached_fetcher.cache_fs
    if fs.exists(filepath):
        fs.remove(filepath)
    soup = bare.get_soup('http://www.google.com')
    assert soup is not None
    assert fs.exists(filepath)


def test_www_errors():
    with pytest.raises(HTTPError):
        bare.get_soup('http://httpstat.us/404')
    with pytest.raises(URLError):
        bare.get_soup('httpd://httpstat.us/404')
    with pytest.raises(HTTPError):
        bare.get_soup('http://httpstat.us/500')



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2015-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The WWW Utils Module (:mod:`tendril.utils.www`)
===============================================

This module provides utilities to deal with the internet. All application code
should access the internet through this module, since this where support for
proxies and caching is implemented.

This module provides three main approaches to handling access to internet
resources :

.. rubric:: urllib based access (:mod:`tendril.utils.www.bare`)

.. autosummary::

    urlopen
    get_soup
    cached_fetcher

.. rubric:: requests based access (:mod:`tendril.utils.www.req`)

.. autosummary::

    get_soup_requests
    get_session

.. rubric:: suds based SOAP access (:mod:`tendril.utils.www.soap`)

.. autosummary::

    get_soap_client

Caching Strategies
------------------

The backends provided by these modules have integrated caching mechanisms
built-in to speed up access to internet based resources.

.. rubric:: Redirect Caching

Redirect caching speeds up network accesses by saving ``301`` and ``302``
redirects, and not needing to get the correct URL on a second access. This
redirect cache is stored as a pickled object in the ``INSTANCE_CACHE``
folder. The effect of this caching is far more apparent when a replicator
cache is also used.

Redirect caching is only supported by the urllib based backend
(:mod:`tendril.utils.www.bare`), and is likely going to be phased out
entirely in the future.

.. rubric:: Full Response Caching

This is a more typical kind of caching, which uses a backend-dependent
mechanism to maintain a cache of full responses received.

.. todo::
    Consider replacing uses of urllib/urllib2 backend with
    :mod:`requests` and simplify this module. Currently, the
    cache provided with the ``requests`` implementation here
    is the major bottleneck and seems to cause a major
    performance hit.

"""

from .helpers import strencode  # noqa

# Bare urllib based client
from .bare import urlopen           # noqa
from .bare import cached_fetcher    # noqa
from .bare import get_soup          # noqa

# Requests based client
from .req import get_session        # noqa
from .req import get_soup_requests  # noqa

# suds based SOAP client
from .soap import get_soap_client   # noqa

# httpx based async client
from .hx import async_client         # noqa
from .hx import with_async_client_cl # noqa